from __future__ import annotations

import sys

import click

from .parseutils import is_destructive
from typing import Any


class ConfirmBoolParamType(click.ParamType):
    name = "confirmation"

    def convert(self, value: bool | str, param: click.Parameter | None, ctx: click.Context | None) -> bool:
        if isinstance(value, bool):
            return value
        value = value.lower()
        if value in ("yes", "y"):
            return True
        if value in ("no", "n"):
            return False
        self.fail(f"{value} is not a valid boolean", param, ctx)

    def __repr__(self) -> str:
        return "BOOL"


BOOLEAN_TYPE = ConfirmBoolParamType()


def confirm_destructive_query(queries: str) -> bool | None:
    """Check if the query is destructive and prompt to confirm.

    Returns:
    - None: non-destructive or cannot prompt (non-tty).
    - True: destructive and user consents.
    - False: destructive and user declines.
    """
    prompt_text = "You're about to run a destructive command.\nDo you want to proceed? (y/n)"
    if is_destructive(queries) and sys.stdin.isatty():
        return bool(prompt(prompt_text, type=BOOLEAN_TYPE))
    return None


def confirm(*args: Any, **kwargs: Any) -> bool:
    """Prompt for confirmation (yes/no) and handle aborts."""
    try:
        return click.confirm(*args, **kwargs)
    except click.Abort:
        return False


def prompt(*args: Any, **kwargs: Any) -> Any:
    """Prompt the user for input and handle aborts. Returns the value from click.prompt."""
    try:
        return click.prompt(*args, **kwargs)
    except click.Abort:
        return False



from __future__ import annotations

import errno
import os
import platform
import shutil
from os.path import dirname, exists, expanduser

from configobj import ConfigObj


def config_location() -> str:
    if "XDG_CONFIG_HOME" in os.environ:
        return "%s/litecli/" % expanduser(os.environ["XDG_CONFIG_HOME"])
    elif platform.system() == "Windows":
        userprofile = os.getenv("USERPROFILE", "")
        return userprofile + "\\AppData\\Local\\dbcli\\litecli\\"
    else:
        return expanduser("~/.config/litecli/")


def load_config(usr_cfg: str, def_cfg: str | None = None) -> ConfigObj:
    cfg = ConfigObj()
    if def_cfg:
        cfg.merge(ConfigObj(def_cfg, interpolation=False))
    cfg.merge(ConfigObj(expanduser(usr_cfg), interpolation=False, encoding="utf-8"))
    cfg.filename = expanduser(usr_cfg)
    return cfg


def ensure_dir_exists(path: str) -> None:
    parent_dir = expanduser(dirname(path))
    try:
        os.makedirs(parent_dir)
    except OSError as exc:
        # ignore existing destination (py2 has no exist_ok arg to makedirs)
        if exc.errno != errno.EEXIST:
            raise


def write_default_config(source: str, destination: str, overwrite: bool = False) -> None:
    destination = expanduser(destination)
    if not overwrite and exists(destination):
        return
    ensure_dir_exists(destination)
    shutil.copyfile(source, destination)


def upgrade_config(config: str, def_config: str) -> None:
    cfg = load_config(config, def_config)
    cfg.write()


def get_config(liteclirc_file: str | None = None) -> ConfigObj:
    from litecli import __file__ as package_root

    package_root = os.path.dirname(str(package_root))

    liteclirc_file = liteclirc_file or f"{config_location()}config"

    default_config = os.path.join(package_root, "liteclirc")
    try:
        write_default_config(default_config, liteclirc_file)
    except OSError:
        # If we can't write to the config file, just use the default config
        return load_config(default_config)

    return load_config(liteclirc_file, default_config)



from __future__ import annotations
import logging
from collections import namedtuple
from enum import Enum
from typing import Any, Callable, cast

from . import export

log = logging.getLogger(__name__)

try:
    import llm  # noqa: F401

    LLM_IMPORTED = True
except ImportError:
    LLM_IMPORTED = False

NO_QUERY = 0
PARSED_QUERY = 1
RAW_QUERY = 2

SpecialCommand = namedtuple(
    "SpecialCommand",
    [
        "handler",
        "command",
        "shortcut",
        "description",
        "arg_type",
        "hidden",
        "case_sensitive",
    ],
)

COMMANDS = {}


@export
class ArgumentMissing(Exception):
    pass


@export
class CommandNotFound(Exception):
    pass


class Verbosity(Enum):
    """Invocation verbosity: succinct (-), normal, or verbose (+)."""

    SUCCINCT = "succinct"
    NORMAL = "normal"
    VERBOSE = "verbose"


@export
def parse_special_command(sql: str) -> tuple[str, "Verbosity", str]:
    """
    Parse a special command, extracting the base command name, verbosity
    (normal, verbose (+), or succinct (-)), and the remaining argument.
    Mirrors the behavior used in similar CLI tools.
    """
    command, _, arg = sql.partition(" ")
    verbosity = Verbosity.NORMAL
    if "+" in command:
        verbosity = Verbosity.VERBOSE
    elif "-" in command:
        verbosity = Verbosity.SUCCINCT
    command = command.strip().strip("+-")
    return (command, verbosity, arg.strip())


@export
def special_command(
    command: str,
    shortcut: str,
    description: str,
    arg_type: int = PARSED_QUERY,
    hidden: bool = False,
    case_sensitive: bool = False,
    aliases: tuple[str, ...] = (),
) -> Callable:
    def wrapper(wrapped: Callable) -> Callable:
        register_special_command(
            wrapped,
            command,
            shortcut,
            description,
            arg_type,
            hidden,
            case_sensitive,
            aliases,
        )
        return wrapped

    return wrapper


@export
def register_special_command(
    handler: Callable,
    command: str,
    shortcut: str,
    description: str,
    arg_type: int = PARSED_QUERY,
    hidden: bool = False,
    case_sensitive: bool = False,
    aliases: tuple[str, ...] = (),
) -> None:
    cmd = command.lower() if not case_sensitive else command
    COMMANDS[cmd] = SpecialCommand(handler, command, shortcut, description, arg_type, hidden, case_sensitive)
    for alias in aliases:
        cmd = alias.lower() if not case_sensitive else alias
        COMMANDS[cmd] = SpecialCommand(
            handler,
            command,
            shortcut,
            description,
            arg_type,
            case_sensitive=case_sensitive,
            hidden=True,
        )


@export
def execute(cur: Any, sql: str) -> list[tuple[Any, ...]]:
    """Execute a special command and return the results. If the special command
    is not supported a KeyError will be raised.
    """
    command, verbosity, arg = parse_special_command(sql)

    if (command not in COMMANDS) and (command.lower() not in COMMANDS):
        raise CommandNotFound

    try:
        special_cmd = COMMANDS[command]
    except KeyError:
        special_cmd = COMMANDS[command.lower()]
        if special_cmd.case_sensitive:
            raise CommandNotFound("Command not found: %s" % command)

    if special_cmd.arg_type == NO_QUERY:
        return cast(list[tuple[Any, ...]], special_cmd.handler())
    elif special_cmd.arg_type == PARSED_QUERY:
        return cast(
            list[tuple[Any, ...]],
            special_cmd.handler(cur=cur, arg=arg, verbose=(verbosity == Verbosity.VERBOSE)),
        )
    elif special_cmd.arg_type == RAW_QUERY:
        return cast(list[tuple[Any, ...]], special_cmd.handler(cur=cur, query=sql))

    raise CommandNotFound(f"Command type not found: {command}")


@special_command("help", "\\?", "Show this help.", arg_type=NO_QUERY, aliases=("\\?", "?"))
def show_help() -> list[tuple]:  # All the parameters are ignored.
    headers = ["Command", "Shortcut", "Description"]
    result = []

    for _, value in sorted(COMMANDS.items()):
        if not value.hidden:
            result.append((value.command, value.shortcut, value.description))
    return [(None, result, headers, None)]


@special_command(".exit", "\\q", "Exit.", arg_type=NO_QUERY, aliases=("\\q", "exit"))
@special_command("quit", "\\q", "Quit.", arg_type=NO_QUERY)
def quit(*_args: Any) -> None:
    raise EOFError


@special_command(
    "\\e",
    "\\e",
    "Edit command with editor (uses $EDITOR).",
    arg_type=NO_QUERY,
    case_sensitive=True,
)
@special_command(
    "\\G",
    "\\G",
    "Display current query results vertically.",
    arg_type=NO_QUERY,
    case_sensitive=True,
)
def stub() -> None:
    raise NotImplementedError


if LLM_IMPORTED:

    @special_command(
        "\\llm",
        "\\ai",
        "Use LLM to construct a SQL query.",
        arg_type=NO_QUERY,
        case_sensitive=False,
        aliases=(".ai", ".llm"),
    )
    def llm_stub() -> None:
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Protocol, Sequence


class DBCursor(Protocol):
    """Minimal DB-API cursor protocol used by special modules."""

    description: Sequence[Sequence[Any]] | None

    # Optional attribute on many DB-API cursors
    connection: Any

    def execute(self, sql: str, params: Any = ...) -> Any: ...

    def fetchall(self) -> list[tuple[Any, ...]]: ...

    def fetchone(self) -> tuple[Any, ...] | None: ...



import os
import shutil
from collections import namedtuple
from datetime import datetime
from textwrap import dedent
from unittest.mock import patch

import click
import pytest
from click.testing import CliRunner
from prompt_toolkit import PromptSession

from litecli.main import LiteCli, cli
from litecli.packages.special.main import COMMANDS as SPECIAL_COMMANDS

from .utils import create_db, db_connection, dbtest, run

test_dir = os.path.abspath(os.path.dirname(__file__))
project_dir = os.path.dirname(test_dir)
default_config_file = os.path.join(project_dir, "tests", "liteclirc")

CLI_ARGS = ["--liteclirc", default_config_file, "_test_db"]

clickoutput: str


@dbtest
def test_execute_arg(executor):
    run(executor, "create table test (a text)")
    run(executor, 'insert into test values("abc")')

    sql = "select * from test;"
    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS + ["-e", sql])

    assert result.exit_code == 0
    assert "abc" in result.output

    result = runner.invoke(cli, args=CLI_ARGS + ["--execute", sql])

    assert result.exit_code == 0
    assert "abc" in result.output

    expected = "a\nabc\n"

    assert expected in result.output


@dbtest
def test_execute_arg_with_table(executor):
    run(executor, "create table test (a text)")
    run(executor, 'insert into test values("abc")')

    sql = "select * from test;"
    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS + ["-e", sql] + ["--table"])
    expected = "+-----+\n| a   |\n+-----+\n| abc |\n+-----+\n"

    assert result.exit_code == 0
    assert expected in result.output


@dbtest
def test_execute_arg_with_csv(executor):
    run(executor, "create table test (a text)")
    run(executor, 'insert into test values("abc")')

    sql = "select * from test;"
    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS + ["-e", sql] + ["--csv"])
    expected = '"a"\n"abc"\n'

    assert result.exit_code == 0
    assert expected in "".join(result.output)


@dbtest
def test_batch_mode(executor):
    run(executor, """create table test(a text)""")
    run(executor, """insert into test values('abc'), ('def'), ('ghi')""")

    sql = "select count(*) from test;\nselect * from test limit 1;"

    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS, input=sql)

    assert result.exit_code == 0
    assert "count(*)\n3\na\nabc\n" in "".join(result.output)


@dbtest
def test_batch_mode_table(executor):
    run(executor, """create table test(a text)""")
    run(executor, """insert into test values('abc'), ('def'), ('ghi')""")

    sql = "select count(*) from test;\nselect * from test limit 1;"

    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS + ["-t"], input=sql)

    expected = dedent(
        """\
        +----------+
        | count(*) |
        +----------+
        | 3        |
        +----------+
        +-----+
        | a   |
        +-----+
        | abc |
        +-----+"""
    )

    assert result.exit_code == 0
    assert expected in result.output


@dbtest
def test_batch_mode_csv(executor):
    run(executor, """create table test(a text, b text)""")
    run(executor, """insert into test (a, b) values('abc', 'de\nf'), ('ghi', 'jkl')""")

    sql = "select * from test;"

    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS + ["--csv"], input=sql)

    expected = '"a","b"\n"abc","de\nf"\n"ghi","jkl"\n'

    assert result.exit_code == 0
    assert expected in "".join(result.output)


def test_help_strings_end_with_periods():
    """Make sure click options have help text that end with a period."""
    for param in cli.params:
        if isinstance(param, click.core.Option):
            assert hasattr(param, "help")
            assert isinstance(param.help, str)
            assert param.help.endswith(".")


def output(monkeypatch, terminal_size, testdata, explicit_pager, expect_pager):
    global clickoutput
    clickoutput = ""
    m = LiteCli(liteclirc=default_config_file)

    class TestOutput:
        def get_size(self):
            size = namedtuple("Size", "rows columns")
            size.columns, size.rows = terminal_size
            return size

    class TestExecute:
        host = "test"
        user = "test"
        dbname = "test"
        port = 0

        def server_type(self):
            return ["test"]

    class PromptBuffer(PromptSession):
        output = TestOutput()

    m.prompt_app = PromptBuffer()
    m.sqlexecute = TestExecute()
    m.explicit_pager = explicit_pager

    def echo_via_pager(s):
        assert expect_pager
        global clickoutput
        clickoutput += s

    def secho(s):
        assert not expect_pager
        global clickoutput
        clickoutput += s + "\n"

    monkeypatch.setattr(click, "echo_via_pager", echo_via_pager)
    monkeypatch.setattr(click, "secho", secho)
    m.output(testdata)
    if clickoutput.endswith("\n"):
        clickoutput = clickoutput[:-1]
    assert clickoutput == "\n".join(testdata)


def test_conditional_pager(monkeypatch):
    testdata = "Lorem ipsum dolor sit amet consectetur adipiscing elit sed do".split(" ")
    # User didn't set pager, output doesn't fit screen -> pager
    output(
        monkeypatch,
        terminal_size=(5, 10),
        testdata=testdata,
        explicit_pager=False,
        expect_pager=True,
    )
    # User didn't set pager, output fits screen -> no pager
    output(
        monkeypatch,
        terminal_size=(20, 20),
        testdata=testdata,
        explicit_pager=False,
        expect_pager=False,
    )
    # User manually configured pager, output doesn't fit screen -> pager
    output(
        monkeypatch,
        terminal_size=(5, 10),
        testdata=testdata,
        explicit_pager=True,
        expect_pager=True,
    )
    # User manually configured pager, output fit screen -> pager
    output(
        monkeypatch,
        terminal_size=(20, 20),
        testdata=testdata,
        explicit_pager=True,
        expect_pager=True,
    )

    SPECIAL_COMMANDS["nopager"].handler()
    output(
        monkeypatch,
        terminal_size=(5, 10),
        testdata=testdata,
        explicit_pager=False,
        expect_pager=False,
    )
    SPECIAL_COMMANDS["pager"].handler("")


def test_reserved_space_is_integer():
    """Make sure that reserved space is returned as an integer."""

    def stub_terminal_size():
        return (5, 5)

    old_func = shutil.get_terminal_size

    shutil.get_terminal_size = stub_terminal_size  # type: ignore[assignment]
    lc = LiteCli()
    assert isinstance(lc.get_reserved_space(), int)
    shutil.get_terminal_size = old_func


@dbtest
def test_import_command(executor):
    data_file = os.path.join(project_dir, "tests", "data", "import_data.csv")
    run(executor, """create table tbl1(one varchar(10), two smallint)""")

    # execute
    run(executor, """.import %s tbl1""" % data_file)

    # verify
    sql = "select * from tbl1;"
    runner = CliRunner()
    result = runner.invoke(cli, args=CLI_ARGS + ["--csv"], input=sql)

    expected = """one","two"
"t1","11"
"t2","22"
"""
    assert result.exit_code == 0
    assert expected in "".join(result.output)


def test_startup_commands(executor):
    m = LiteCli(liteclirc=default_config_file)
    assert m.startup_commands
    assert m.startup_commands["commands"] == [
        "create table startupcommands(a text)",
        "insert into startupcommands values('abc')",
    ]

    # implement tests on executions of the startupcommands


@patch("litecli.main.datetime")  # Adjust if your module path is different
def test_get_prompt(mock_datetime):
    # We'll freeze time at 2025-01-20 13:37:42 for comedic effect.
    # Because "leet" times call for 13:37!
    frozen_time = datetime(2025, 1, 20, 13, 37, 42)
    mock_datetime.now.return_value = frozen_time
    # Ensure `datetime` class is still accessible for strftime usage
    mock_datetime.datetime = datetime

    # Instantiate and connect
    lc = LiteCli()
    lc.connect("/tmp/litecli_test.db")

    # 1. Test \d => full path to the DB
    assert lc.get_prompt(r"\d") == "/tmp/litecli_test.db"

    # 2. Test \f => basename of the DB
    #    (because "f" stands for "filename", presumably!)
    assert lc.get_prompt(r"\f") == "litecli_test.db"

    # 3. Test \_ => single space
    assert lc.get_prompt(r"Hello\_World") == "Hello World"

    # 4. Test \n => newline
    #    Just to be sure we're only inserting a newline,
    #    we can check length or assert the presence of "\n".
    expected = f"Line1{os.linesep}Line2"
    assert lc.get_prompt(r"Line1\nLine2") == expected

    # 5. Test date/time placeholders (with frozen time):
    #    \D => e.g. 'Mon Jan 20 13:37:42 2025'
    expected_date_str = frozen_time.strftime("%a %b %d %H:%M:%S %Y")
    assert lc.get_prompt(r"\D") == expected_date_str

    # 6. Test \m => minutes
    assert lc.get_prompt(r"\m") == "37"

    # 7. Test \P => AM/PM
    #    13:37 is PM
    assert lc.get_prompt(r"\P") == "PM"

    # 8. Test \R => 24-hour format hour
    assert lc.get_prompt(r"\R") == "13"

    # 9. Test \r => 12-hour format hour
    #    13:37 is 01 in 12-hour format
    assert lc.get_prompt(r"\r") == "01"

    # 10. Test \s => seconds
    assert lc.get_prompt(r"\s") == "42"

    # 11. Test when dbname is None => (none)
    lc.connect(None)
    # Simulate no DB connection and incorrect argument type
    assert lc.get_prompt(r"\d") == "(none)"
    assert lc.get_prompt(r"\f") == "(none)"

    # 12. Windows path
    lc.connect("C:\\Users\\litecli\\litecli_test.db")
    assert lc.get_prompt(r"\d") == "C:\\Users\\litecli\\litecli_test.db"


@pytest.mark.parametrize(
    "uri, expected_dbname",
    [
        ("file:{tmp_path}/test.db", "{tmp_path}/test.db"),
        ("file:{tmp_path}/test.db?mode=ro", "{tmp_path}/test.db"),
        ("file:{tmp_path}/test.db?mode=ro&cache=shared", "{tmp_path}/test.db"),
    ],
)
def test_file_uri(tmp_path, uri, expected_dbname):
    """
    Test that `file:` URIs are correctly handled
    ref:
        https://docs.python.org/3/library/sqlite3.html#sqlite3-uri-tricks
        https://www.sqlite.org/c3ref/open.html#urifilenameexamples
    """
    # - ensure db exists
    db_path = tmp_path / "test.db"
    create_db(db_path)
    db_connection(db_path)
    uri = uri.format(tmp_path=tmp_path)

    lc = LiteCli()
    lc.connect(uri)

    assert lc.get_prompt(r"\d") == expected_dbname.format(tmp_path=tmp_path)



# ruff: noqa

from __future__ import annotations
from types import FunctionType

from typing import Callable, Any

__all__: list[str] = []


def export(defn: Callable[..., Any]) -> Callable[..., Any]:
    """Decorator to explicitly mark functions that are exposed in a lib."""
    # ty, requires explict check for callable of tyep | function type to access __name__
    if isinstance(defn, (type, FunctionType)):
        globals()[defn.__name__] = defn
        __all__.append(defn.__name__)
    return defn


from . import dbcommands
from . import iocommands
from . import llm
from . import utils
from .main import CommandNotFound, register_special_command, execute
from .iocommands import (
    set_favorite_queries,
    editor_command,
    get_filename,
    get_editor_query,
    open_external_editor,
    is_expanded_output,
    set_expanded_output,
    write_tee,
    unset_once_if_written,
    unset_pipe_once_if_written,
    disable_pager,
    set_pager,
    is_pager_enabled,
    write_once,
    write_pipe_once,
    close_tee,
)
from .llm import is_llm_command, handle_llm, FinishIteration



from __future__ import annotations

from pygments.lexer import inherit
from pygments.lexers.sql import MySqlLexer
from pygments.token import Keyword


class LiteCliLexer(MySqlLexer):
    """Extends SQLite lexer to add keywords."""

    tokens = {
        "root": [
            (r"\brepair\b", Keyword),
            (r"\boffset\b", Keyword),
            inherit,
        ]
    }



from __future__ import annotations
import logging
from typing import Any

from prompt_toolkit.enums import EditingMode
from prompt_toolkit.filters import completion_is_selected
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.key_processor import KeyPressEvent

_logger = logging.getLogger(__name__)


def cli_bindings(cli: Any) -> KeyBindings:
    """Custom key bindings for cli."""
    kb = KeyBindings()

    @kb.add("f3")
    def _(_event: KeyPressEvent) -> None:
        """Enable/Disable Multiline Mode."""
        _logger.debug("Detected F3 key.")
        cli.multi_line = not cli.multi_line

    @kb.add("f4")
    def _(event: KeyPressEvent) -> None:
        """Toggle between Vi and Emacs mode."""
        _logger.debug("Detected F4 key.")
        if cli.key_bindings == "vi":
            event.app.editing_mode = EditingMode.EMACS
            cli.key_bindings = "emacs"
        else:
            event.app.editing_mode = EditingMode.VI
            cli.key_bindings = "vi"

    @kb.add("tab")
    def _(event: KeyPressEvent) -> None:
        """Force autocompletion at cursor."""
        _logger.debug("Detected <Tab> key.")
        b = event.app.current_buffer
        if b.complete_state:
            b.complete_next()
        else:
            b.start_completion(select_first=True)

    @kb.add("s-tab")
    def _(event: KeyPressEvent) -> None:
        """Force autocompletion at cursor."""
        _logger.debug("Detected <Tab> key.")
        b = event.app.current_buffer
        if b.complete_state:
            b.complete_previous()
        else:
            b.start_completion(select_last=True)

    @kb.add("c-space")
    def _(event: KeyPressEvent) -> None:
        """
        Initialize autocompletion at cursor.

        If the autocompletion menu is not showing, display it with the
        appropriate completions for the context.

        If the menu is showing, select the next completion.
        """
        _logger.debug("Detected <C-Space> key.")

        b = event.app.current_buffer
        if b.complete_state:
            b.complete_next()
        else:
            b.start_completion(select_first=False)

    @kb.add("enter", filter=completion_is_selected)
    def _(event: KeyPressEvent) -> None:
        """Makes the enter key work as the tab key only when showing the menu.

        In other words, don't execute query when enter is pressed in
        the completion dropdown menu, instead close the dropdown menu
        (accept current selection).

        """
        _logger.debug("Detected enter key.")

        event.current_buffer.complete_state = None
        b = event.app.current_buffer
        b.complete_state = None

    @kb.add("right", filter=completion_is_selected)
    def _(event: KeyPressEvent) -> None:
        """Accept the completion that is selected in the dropdown menu."""
        _logger.debug("Detected right-arrow key.")

        b = event.app.current_buffer
        b.complete_state = None

    return kb
